@@ -31,6 +31,7 @@
 #include "command_adaptation.h"
 #include "log.h"
 #include "config.h"
+#include "../cli/hash.h"
 
 struct Command_ATSHA204 make_command ()
 {
@@ -319,7 +320,7 @@ bool write4 (int fd, enum DATA_ZONE zone, uint8_t addr, uint32_t buf)
   set_data (&c, (uint8_t *)&buf, sizeof (buf));
   set_execution_time (&c, 0, 4000000);
 
-  if (RSP_SUCCESS == process_command (fd, &c, &recv, sizeof (recv)));
+  if (RSP_SUCCESS == process_command (fd, &c, &recv, sizeof (recv)))
   {
     if (0 == (int) recv)
       status = true;
@@ -332,17 +333,30 @@ bool write4 (int fd, enum DATA_ZONE zone, uint8_t addr, uint32_t buf)
 }
 
 bool write32 (int fd, enum DATA_ZONE zone, uint8_t addr,
-              struct octet_buffer buf)
+              struct octet_buffer buf, struct octet_buffer *mac)
 {
 
   assert (NULL != buf.ptr);
   assert (32 == buf.len);
+  if (NULL != mac)
+    assert (NULL != mac->ptr);
 
   bool status = false;
   uint8_t recv = 0;
   uint8_t param2[2] = {0};
   uint8_t param1 = set_zone_bits (zone);
 
+  struct octet_buffer data = {0,0};
+
+  if (NULL != mac)
+    data = make_buffer (buf.len + mac->len);
+  else
+    data = make_buffer (buf.len);
+
+  memcpy (data.ptr, buf.ptr, buf.len);
+  if (NULL != mac && mac->len > 0)
+    memcpy (data.ptr + buf.len, mac->ptr, mac->len);
+
   /* If writing 32 bytes, this bit must be set in param1 */
   uint8_t WRITE_32_MASK = 0b10000000;
 
@@ -355,75 +369,25 @@ bool write32 (int fd, enum DATA_ZONE zone, uint8_t addr,
   set_opcode (&c, COMMAND_WRITE);
   set_param1 (&c, param1);
   set_param2 (&c, param2);
-  set_data (&c, buf.ptr, buf.len);
+  set_data (&c, data.ptr, data.len);
+
   set_execution_time (&c, 0, WRITE_AVG_EXEC);
 
-  if (RSP_SUCCESS == process_command (fd, &c, &recv, sizeof (recv)));
+  if (RSP_SUCCESS == process_command (fd, &c, &recv, sizeof (recv)))
   {
+    CTX_LOG (DEBUG, "Write 32 successful.");
     if (0 == (int) recv)
       status = true;
   }
 
+  free_octet_buffer (data);
+
   return status;
 
 
 
 }
 
-struct octet_buffer gen_nonce (int fd, int seed_update_flag,
-                              struct octet_buffer input)
-
-{
-
-  uint8_t *recv = NULL;
-  uint8_t param1 = seed_update_flag;
-  uint8_t param2[2] = {0};
-  unsigned int recv_len = 0;
-  struct octet_buffer response = {NULL, 0};
-
-  assert (1 == seed_update_flag || 0 == seed_update_flag);
-  assert (NULL != input.ptr);
-  /* If 32, the nonce is considered a pass through and will be used
-     directly by the system */
-  /* If 20, the nonce will be combined with a random number */
-  assert (32 == input.len || 20 == input.len);
-
-  if (32 == input.len)
-    {
-      recv_len = 1;
-    }
-  else
-    {
-      recv_len = 32;
-    }
-
-  recv = malloc (recv_len);
-  assert (NULL != recv);
-
-
-  struct Command_ATSHA204 c = make_command ();
-
-  set_opcode (&c, COMMAND_NONCE);
-  set_param1 (&c, param1);
-  set_param2 (&c, param2);
-  set_data (&c, input.ptr, input.len);
-  set_execution_time (&c, 0, 22000000); /* avg. 22 msec */
-
-  if (RSP_SUCCESS == process_command (fd, &c, recv, recv_len));
-  {
-    response.ptr = recv;
-    response.len= recv_len;
-  }
-
-  return response;
-
-
-
-}
-
-
-
-
 uint8_t serialize_check_mac_mode (struct check_mac_encoding c)
 {
   /* The serialized result */
@@ -849,20 +813,22 @@ bool set_otp_zone (int fd, struct octet_buffer *otp_zone)
   buf.len = sizeof (nulls);
 
   /* Fill the OTP zone with blanks from their default FFFF */
-  success = write32 (fd, OTP_ZONE, 0, buf);
+  success = write32 (fd, OTP_ZONE, 0, buf, NULL);
 
   if (success)
-    success = write32 (fd, OTP_ZONE, SIZE_OF_WRITE / sizeof (uint32_t), buf);
+    success = write32 (fd, OTP_ZONE, SIZE_OF_WRITE / sizeof (uint32_t),
+                       buf, NULL);
 
   /* Fill in the data */
   buf.ptr = part1;
   CTX_LOG (DEBUG, "Writing: %s", buf.ptr);
   if (success)
-    success = write32 (fd, OTP_ZONE, 0, buf);
+    success = write32 (fd, OTP_ZONE, 0, buf, NULL);
   buf.ptr = part2;
   CTX_LOG (DEBUG, "Writing: %s", buf.ptr);
   if (success)
-    success = write32 (fd, OTP_ZONE, SIZE_OF_WRITE / sizeof (uint32_t), buf);
+    success = write32 (fd, OTP_ZONE, SIZE_OF_WRITE / sizeof (uint32_t),
+                       buf, NULL);
 
   /* Lastly, copy the OTP zone into one contiguous buffer.
      Ironically, the OTP can't be read while unlocked. */
@@ -927,4 +893,343 @@ enum DEVICE_STATE get_device_state (int fd)
 
   return state;
 
+}
+
+uint8_t slot_to_addr (enum DATA_ZONE zone, uint8_t slot)
+{
+    switch (zone)
+      {
+      case DATA_ZONE:
+        assert (0 <= slot && slot <= 15);
+        break;
+
+      case OTP_ZONE:
+        assert (0 == slot || 1 == slot);
+        break;
+
+      case CONFIG_ZONE:
+        assert (0 <= slot && slot <= 2);
+        break;
+
+      default:
+        assert (false);
+      }
+
+    slot <<= 3;
+
+    return slot;
+
+}
+
+struct octet_buffer gen_nonce (int fd, struct octet_buffer data)
+{
+  const unsigned int EXTERNAL_INPUT_LEN = 32;
+  const unsigned int NEW_NONCE_LEN = 20;
+
+  assert (NULL != data.ptr && (EXTERNAL_INPUT_LEN == data.len ||
+                               NEW_NONCE_LEN == data.len));
+
+  uint8_t param2[2] = {0};
+  uint8_t param1 = 0;
+
+  unsigned int rsp_len = 0;
+
+  if (EXTERNAL_INPUT_LEN == data.len)
+    {
+      const unsigned int PASS_THROUGH_MODE = 3;
+      const unsigned int RSP_LENGTH = 1;
+      param1 = PASS_THROUGH_MODE;
+      rsp_len = RSP_LENGTH;
+    }
+  else
+    {
+      const unsigned int COMBINE_AND_UPDATE_SEED = 0;
+      const unsigned int RSP_LENGTH = 32;
+      param1 = COMBINE_AND_UPDATE_SEED;
+      rsp_len = RSP_LENGTH;
+    }
+
+  struct octet_buffer buf = make_buffer (rsp_len);
+
+  struct Command_ATSHA204 c = make_command ();
+
+  set_opcode (&c, COMMAND_NONCE);
+  set_param1 (&c, param1);
+  set_param2 (&c, param2);
+  set_data (&c, data.ptr, data.len);
+  set_execution_time (&c, 0, NONCE_AVG_EXEC);
+
+  if (RSP_SUCCESS != process_command (fd, &c, buf.ptr, buf.len))
+    {
+      CTX_LOG (DEBUG, "Nonce command failed");
+      free_octet_buffer (buf);
+      buf.ptr = NULL;
+    }
+
+  return buf;
+
+
+
+}
+
+struct octet_buffer get_nonce (int fd)
+{
+  struct octet_buffer otp;
+  struct octet_buffer nonce = {0, 0};
+  const unsigned int MIX_DATA_LEN = 20;
+
+  otp = get_otp_zone (fd);
+  unsigned int otp_len = otp.len;
+
+  if (otp.len > MIX_DATA_LEN && otp.ptr != NULL)
+    {
+      otp.len = MIX_DATA_LEN;
+      nonce = gen_nonce (fd, otp);
+      otp.len = otp_len;
+
+    }
+
+  free_octet_buffer (otp);
+
+  return nonce;
+}
+
+
+struct octet_buffer gen_temp_key_from_nonce (int fd, struct octet_buffer random,
+                                             const struct octet_buffer otp)
+{
+  assert (NULL != random.ptr && 32 == random.len);
+
+  struct octet_buffer result = {0,0};
+
+  const unsigned int MIX_DATA_LEN = 20;
+  const uint8_t OPCODE = 0x16;
+  const uint8_t MODE = 0;
+  const uint8_t PARAM2 = 0; /* only the LSB are used */
+
+  const unsigned int data_len = random.len + MIX_DATA_LEN +
+    sizeof (OPCODE) + sizeof (MODE) + sizeof (PARAM2);
+
+
+  if (otp.len > MIX_DATA_LEN && otp.ptr != NULL)
+    {
+      struct octet_buffer data_to_hash = make_buffer (data_len);
+
+      unsigned int offset = 0;
+
+      offset = copy_buffer (data_to_hash, offset, random);
+
+      offset = copy_to_buffer (data_to_hash, offset, otp.ptr, MIX_DATA_LEN);
+
+      data_to_hash.ptr[offset] = OPCODE;
+      offset += 1;
+
+      data_to_hash.ptr[offset] = MODE;
+      offset += 1;
+
+      data_to_hash.ptr[offset] = PARAM2;
+      offset += 1;
+
+      assert (offset == data_len);
+
+
+      print_hex_string ("Nonce data to hash", data_to_hash.ptr,
+                        data_to_hash.len);
+
+
+      result = sha256_buffer (data_to_hash);
+
+      print_hex_string ("Nonce temp key", result.ptr,
+                        result.len);
+    }
+
+  return result;
+
+}
+
+struct octet_buffer gen_temp_key_from_digest (int fd,
+                                              struct octet_buffer prev_temp_key,
+                                              unsigned int slot,
+                                              struct octet_buffer key)
+{
+  assert (NULL != prev_temp_key.ptr && 32 == prev_temp_key.len);
+  assert (NULL != key.ptr && 32 == key.len);
+  assert (slot <= 15);
+
+  struct octet_buffer result = {0,0};
+
+  const uint8_t OPCODE = COMMAND_GEN_DIG;
+  const uint8_t PARAM1 = 0x02;
+  uint8_t PARAM2[2] = {0};
+  PARAM2[1] = slot;
+
+  const uint8_t sn8 = 0xEE;
+  const uint8_t sn0 = 0x01;
+  const uint8_t sn1 = 0x23;
+  const unsigned int ZERO_25 = 25;
+
+  unsigned int len = key.len + sizeof (OPCODE) + sizeof (PARAM1) +
+    sizeof (PARAM2) + sizeof (sn8) + sizeof (sn0) + sizeof (sn1) + ZERO_25 +
+    prev_temp_key.len;
+
+  struct octet_buffer data_to_hash = make_buffer (len);
+
+  unsigned int offset = 0;
+
+  offset = copy_buffer (data_to_hash, offset, key);
+
+  data_to_hash.ptr[offset] = OPCODE;
+  offset += 1;
+
+  data_to_hash.ptr[offset] = PARAM1;
+  offset += 1;
+
+  offset = copy_to_buffer (data_to_hash, offset, PARAM2, 2);
+
+  data_to_hash.ptr[offset] = sn8;
+  offset += 1;
+
+  data_to_hash.ptr[offset] = sn0;
+  offset += 1;
+
+  data_to_hash.ptr[offset] = sn1;
+  offset += 1;
+
+  struct octet_buffer zeros = make_buffer (ZERO_25);
+
+  offset = copy_buffer (data_to_hash, offset, zeros);
+
+  offset = copy_buffer (data_to_hash, offset, prev_temp_key);
+
+  assert (offset == len);
+
+  print_hex_string ("Temp Key data to hash", data_to_hash.ptr, data_to_hash.len);
+
+  result = sha256_buffer (data_to_hash);
+
+  print_hex_string ("Temp Key", result.ptr, result.len);
+
+  free_octet_buffer (zeros);
+
+  return result;
+
+
+
+
+
+}
+
+struct octet_buffer mac_write (const struct octet_buffer temp_key,
+                               uint8_t opcode,
+                               uint8_t param1, const uint8_t *param2,
+                               const struct octet_buffer data)
+{
+
+  assert (NULL != param2);
+  assert (NULL != temp_key.ptr && temp_key.len == 32);
+  assert (NULL != data.ptr && data.len == 32);
+
+  const uint8_t sn8 = 0xEE;
+  const uint8_t sn0 = 0x01;
+  const uint8_t sn1 = 0x23;
+  const unsigned int ZERO_25 = 25;
+
+  unsigned int len = temp_key.len + sizeof (opcode) + sizeof (param1) + 2 +
+    sizeof (sn8) + sizeof (sn0) + sizeof (sn1) + ZERO_25 + data.len;
+
+  struct octet_buffer data_to_hash = make_buffer (len);
+
+  unsigned int offset = 0;
+
+  offset = copy_buffer (data_to_hash, offset, temp_key);
+
+  data_to_hash.ptr[offset] = opcode;
+  offset += 1;
+
+  data_to_hash.ptr[offset] = param1;
+  offset += 1;
+
+  offset = copy_to_buffer (data_to_hash, offset, param2, 2);
+
+  data_to_hash.ptr[offset] = sn8;
+  offset += 1;
+
+  data_to_hash.ptr[offset] = sn0;
+  offset += 1;
+
+  data_to_hash.ptr[offset] = sn1;
+  offset += 1;
+
+  struct octet_buffer zeros = make_buffer (ZERO_25);
+
+  offset = copy_buffer (data_to_hash, offset, zeros);
+
+  offset = copy_buffer (data_to_hash, offset, data);
+
+  assert (offset == len);
+
+  print_hex_string ("Write mac data to hash", data_to_hash.ptr,
+                    data_to_hash.len);
+
+  struct octet_buffer mac = sha256_buffer (data_to_hash);
+
+  free_octet_buffer (zeros);
+
+  print_hex_string ("Mac'd write", mac.ptr, mac.len);
+
+  return mac;
+
+}
+
+bool gen_digest (int fd, enum DATA_ZONE zone, unsigned int slot)
+{
+
+  bool result = false;
+
+  if (DATA_ZONE == zone)
+    assert (slot <= 15);
+  else
+    assert (slot <= 1);
+
+  uint8_t param2[2] = {0};
+  uint8_t param1 = 0;
+
+  switch (zone)
+    {
+    case DATA_ZONE:
+      param1 = 0x02;
+      break;
+    case OTP_ZONE:
+      param1 = 0x01;
+      break;
+    case CONFIG_ZONE:
+      param1 = 0x00;
+      break;
+    default:
+      assert (false);
+    }
+
+  param2[1] = slot;
+
+
+  uint8_t rsp = 0;
+
+  struct Command_ATSHA204 c = make_command ();
+
+  set_opcode (&c, COMMAND_GEN_DIG);
+  set_param1 (&c, param1);
+  set_param2 (&c, param2);
+  set_data (&c, NULL, 0);
+  set_execution_time (&c, 0, GEN_DIG_AVG_EXEC);
+
+  if (RSP_SUCCESS == process_command (fd, &c, &rsp, sizeof (rsp)))
+    {
+      result = true;
+    }
+
+  return result;
+
+
+
+
 }
